"""
    QApp Platform Project
    pennylane_provider_factory.py
    Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from quapp_common.config.logging_config import logger
from quapp_common.enum.provider_tag import ProviderTag
from quapp_common.enum.sdk import Sdk
from quapp_common.factory.provider_factory import ProviderFactory

from ..model.provider.rigetti_pyquil_provider import RigettiPyquilProvider
from ..model.provider.qapp_pyquil_provider import QAppPyquilProvider

class PyquilProviderFactory(ProviderFactory):

    @staticmethod
    def create_provider(provider_type: ProviderTag, sdk: Sdk, authentication: dict):
        logger.info("[PyquilProviderFactory] create_provider()")
        logger.debug(f"provider_type: {provider_type}, sdk: {sdk}, authentication: {authentication}")

        match provider_type:
            case ProviderTag.QUAO_QUANTUM_SIMULATOR:
                if Sdk.PYQUIL.__eq__(sdk):
                    return QAppPyquilProvider()
                raise ValueError(f'Unsupported SDK for provider type: {provider_type}')
            case ProviderTag.RIGETTI:
                if Sdk.PYQUIL.__eq__(sdk):
                    return RigettiPyquilProvider(
                        authentication.get("deviceName"),
                        authentication.get("clientId"),
                        authentication.get("issuer"),
                        authentication.get("refreshToken"),
                        authentication.get("accessToken"),
                        authentication.get("apiUrl")
                    )
                raise ValueError(f'Unsupported SDK for provider type: {provider_type}')

            case _:
                raise ValueError(f'Unsupported provider type: {provider_type}')


"""
    QApp Platform Project pennylane_invocation.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""

from quapp_common.component.backend.invocation import Invocation
from quapp_common.config.logging_config import logger
from quapp_common.config.thread_config import circuit_exporting_pool
from quapp_common.data.async_task.circuit_export.backend_holder import BackendDataHolder
from quapp_common.data.async_task.circuit_export.circuit_holder import CircuitDataHolder
from quapp_common.data.request.invocation_request import InvocationRequest
from quapp_common.model.provider.provider import Provider

from ...async_tasks.pyquil_circuit_export_task import PyquilCircuitExportTask
from ...factory.pyquil_device_factory import PyquilDeviceFactory
from ...factory.pyquil_provider_factory import PyquilProviderFactory

class PyquilInvocation(Invocation):

    def __init__(self, request_data: InvocationRequest, **kwargs):
        super().__init__(request_data)
        self.num_qubits = kwargs.get('num_qubits')

    def _export_circuit(self, circuit):
        logger.info("[PyquilInvocation] _export_circuit()")

        circuit_export_task = PyquilCircuitExportTask(
            circuit_data_holder=CircuitDataHolder(circuit, self.circuit_export_url),
            backend_data_holder=BackendDataHolder(
                self.backend_information, self.authentication.user_token
            ),
        )
        circuit_exporting_pool.submit(circuit_export_task.do)

    def _create_provider(self):
        logger.info('[PennylaneInvocation] _create_provider()')

        return PyquilProviderFactory.create_provider(
            provider_type=self.backend_information.provider_tag, sdk=self.sdk,
            authentication=self.backend_information.authentication)

    def _create_device(self, provider: Provider):
        logger.info('[PennylaneInvocation] Creating device')
        return PyquilDeviceFactory.create_device(provider=provider,
                                                    device_specification=self.backend_information.device_name,
                                                    authentication=self.backend_information.authentication,
                                                    sdk=self.sdk, num_qubits=self.num_qubits,
                                                    input=self.input)

    def _get_qubit_amount(self, circuit):
        return len(circuit.get_qubit_indices())


"""
    QApp Platform Project qiskit_job_fetching.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from quapp_common.component.backend.job_fetcher import JobFetcher
from quapp_common.data.request.job_fetching_request import JobFetchingRequest
from quapp_common.config.logging_config import logger
from quapp_common.enum.provider_tag import ProviderTag
from quapp_common.enum.sdk import Sdk
from quapp_common.enum.status.job_status import JobStatus
from quapp_common.enum.status.status_code import StatusCode
from pyquil import get_qc
import pickle
import os

from ...factory.pyquil_provider_factory import PyquilProviderFactory


class PyquilJobFetching(JobFetcher):

    def __init__(self, request_data: JobFetchingRequest, device_name: str, authentication: dict):
        super().__init__(request_data)
        self.request_data = request_data
        self.device_name = device_name
        self.authentication = authentication

    def _collect_provider(self):

        device = PyquilProviderFactory.create_provider(authentication=self.authentication,
                                              sdk=Sdk.PYQUIL,
                                              provider_type=ProviderTag.RIGETTI)

        print(f"Device name in fetching: {self.device_name}")
        return device.get_backend(self.device_name)
        

    def _retrieve_job(self, job):
        
        job_id = self.provider_job_id
        device = self._collect_provider()

        save_dir = "/home/app"

        file_path = os.path.join(save_dir, f"{job_id}.pkl")

        with open(file_path, 'rb') as file:
            data = pickle.load(file)
        
        res = device.qam.get_result(data)
        print(f"Result: {res}")

        return res

    def _get_job_status(self, job):

        if (job.execution_duration_microseconds is not None):
            
            if os.path.exists(self.provider_job_id+".pkl"):
                os.remove(self.provider_job_id+".pkl")
            return JobStatus.DONE.value

        return "POLLING"

    def _get_job_result(self, job):
        return job

    def _produce_histogram_data(self, job_result) -> dict | None:
        return None

    def _get_execution_time(self, job_result):
        return job_result.execution_duration_microseconds
        
    def _get_shots(self, job_result):
        return None



"""
    QApp Platform Project
    pennylane_device_factory.py
    Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from quapp_common.config.logging_config import logger
from quapp_common.enum.provider_tag import ProviderTag
from quapp_common.enum.sdk import Sdk
from quapp_common.factory.device_factory import DeviceFactory
from quapp_common.model.provider.provider import Provider
from ..model.device.qapp_pyquil_device import QAppPyquilDevice
from ..model.device.rigetti_pyquil_device import RigettiPyquilDevice


class PyquilDeviceFactory(DeviceFactory):

    @staticmethod
    def create_device(provider: Provider, device_specification: str, authentication: dict, sdk: Sdk,
                      **kwargs):
        logger.info("[PyquilDeviceFactory] create_device()")

        provider_type = ProviderTag.resolve(provider.get_provider_type().value)

        logger.info("[PyquilDeviceFactory] provider type:" + str(provider_type))

        match provider_type:
            case ProviderTag.QUAO_QUANTUM_SIMULATOR:
                if Sdk.PYQUIL == sdk:
                    logger.debug('[PyquilDeviceFactory] Creating QAppPyquilDevice')
                    return QAppPyquilDevice(provider, device_specification)
            case ProviderTag.RIGETTI:
                if Sdk.PYQUIL == sdk:
                    logger.debug('[PyquilDeviceFactory] Creating RigettiPyquilDevice')
                    return RigettiPyquilDevice(provider, device_specification)
            case _:
                raise ValueError(f"Unsupported provider type: {provider_type}")


"""
    QApp Platform Project
    qapp_pyquil_device.py
    Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
import time
import math

from quapp_common.data.response.authentication import Authentication
from quapp_common.data.response.project_header import ProjectHeader
from quapp_common.model.device.device import Device
from quapp_common.config.logging_config import logger
from quapp_common.data.device.circuit_running_option import CircuitRunningOption
from quapp_common.model.provider.provider import Provider
from quapp_common.enum.invocation_step import InvocationStep


class QAppPyquilDevice(Device):
    def __init__(self, provider: Provider, device_specification: str):
        super().__init__(provider, device_specification)
        logger.debug('[QAppPyquilDevice] Initializing device specification')
        self.device_specification = device_specification

    def _create_job(self, circuit, options: CircuitRunningOption):
        logger.debug(
            '[QAppPyquilDevice] Creating job with {0} shots'.format(
                options.shots))
        
        circuit.wrap_in_numshots_loop(options.shots)
      
        start_time = time.time()
        executable = self.device.compile(circuit)
        result = self.device.run(executable)
        end_time = time.time()

        data = {"result": result, "time_taken_execute": end_time - start_time}

        # logger.info(data)
        return data

    def _is_simulator(self) -> bool:
        logger.debug('[QAppPyquilDevice] Is simulator')
        return True

    def _produce_histogram_data(self, job_result) -> dict | None:
        logger.info('[PyquilDevice] Producing histogram data')

        histogram = job_result.get('histogram')

        if histogram is None:
            logger.debug("[PyquilDevice] Can't produce histogram")

        return job_result.get('histogram')

    def _get_provider_job_id(self, job) -> str:
        logger.debug('[PyquilDevice] Getting job id')

        # no job id in local simulator
        return ""

    def _get_job_status(self, job) -> str:
        logger.debug('[PyquilDevice] Getting job status')

        return "DONE"

    def _get_job_result(self, job) -> dict:
        logger.debug('[PyquilDevice] Getting job result')

        return job

    def _calculate_execution_time(self, job_result):
        logger.debug('[PyquilDevice] Calculating execution time')

        self.execution_time = job_result.get('time_taken_execute')

        logger.debug(
            '[PyquilDevice] Execution time calculation was: {0} seconds'.format(
                self.execution_time))

    def run_circuit(self,
                    circuit,
                    post_processing_fn,
                    options: CircuitRunningOption,
                    callback_dict: dict,
                    authentication: Authentication,
                    project_header: ProjectHeader):
        """
        @param project_header: project header
        @param callback_dict: callback url dictionary
        @param options: Options for run circuit
        @param authentication: Authentication for calling quao server
        @param post_processing_fn: Post-processing function
        @param circuit: Circuit was run
        """
        original_job_result, job_response = self._on_execution(
            authentication=authentication,
            project_header=project_header,
            execution_callback=callback_dict.get(InvocationStep.EXECUTION),
            circuit=circuit,
            options=options)

        if original_job_result is None:
            return

        job_response = self._on_analysis(
            job_response=job_response,
            original_job_result=original_job_result,
            analysis_callback=callback_dict.get(InvocationStep.ANALYSIS))

        if job_response is None:
            return

        self._on_finalization(job_result=original_job_result.get('result'),
                              authentication=authentication,
                              post_processing_fn=post_processing_fn,
                              finalization_callback=callback_dict.get(
                                  InvocationStep.FINALIZATION),
                              project_header=project_header)



"""
    QApp Platform Project
    qapp_pennylane_provider.py
    Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from quapp_common.config.logging_config import logger
from quapp_common.enum.provider_tag import ProviderTag
from quapp_common.model.provider.provider import Provider

from pyquil import get_qc


class QAppPyquilProvider(Provider):
    def __init__(self, ):
        logger.debug('[QAppPyquilProvider] get_backend()')
        super().__init__(ProviderTag.QUAO_QUANTUM_SIMULATOR)

    def get_backend(self, device_specification):
        logger.debug('[QAppPyquilProvider] get_backend()')

        try:
            print(device_specification)
            return get_qc(device_specification,as_qvm=True)
        except Exception as e:
            print(e)
            raise ValueError('[QAppPyquilProvider] Unsupported device')

    def collect_provider(self):
        logger.debug('[QAppPyquilProvider] collect_provider()')
        return None


"""
    QApp Platform Project pyquil_handler_factory.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""

from quapp_common.config.logging_config import logger
from quapp_common.factory.handler_factory import HandlerFactory
from quapp_common.handler.handler import Handler

from ..handler.invocation_handler import InvocationHandler
from ..handler.job_fetching_handler import JobFetchingHandler

class PyquilHandlerFactory(HandlerFactory):

    @staticmethod
    def create_handler(event, circuit_preparation_fn, post_processing_fn) -> Handler:
        logger.info("[PyquilHandlerFactory] create_handler()")

        request_data = event.json()
        provider_job_id = request_data.get("providerJobId")

        if provider_job_id is None:
            logger.debug("[PyquilHandlerFactory] Create InvocationHandler")
            return InvocationHandler(
                request_data=request_data,
                circuit_preparation_fn=circuit_preparation_fn,
                post_processing_fn=post_processing_fn,
            )
        # logger.debug("Create JobFetchingHandler")
        # return JobFetchingHandler(
        #     request_data=request_data, post_processing_fn=post_processing_fn
        # )



"""
    QApp Platform Project pennylane_circuit_export_task.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from quapp_common.async_tasks.export_circuit_task import CircuitExportTask
# from ..quapp_common.config.logging_config import logger
# from pyquil.latex import to_latex
# import os
# import subprocess


class PyquilCircuitExportTask(CircuitExportTask):
        
    def _convert(self):
        
        # circuit = self.circuit_data_holder.circuit

        # latex_code = to_latex(circuit)

        # latex_filename = "quantikz.tex"
        # pdf_filename = "quantikz.pdf"
        # svg_filename = "quantikz.svg"

        
        # try:

        #     with open(latex_filename, "w") as latex_file:
        #         latex_file.write(latex_code)

        #     subprocess.run(["pdflatex", latex_filename], check=True)
        #     subprocess.run(["pdf2svg", pdf_filename, svg_filename], check=True)

        #     with open(svg_filename, "rb") as svg_file:
        #         svg_bytes = svg_file.read()

        # except subprocess.CalledProcessError:
        #     logger.error('[PyquilCircuitExportTask] Fail to convert to svg')
        #     return None
        
        # finally:
            
        #     logger.info('[PyquilCircuitExportTask] Cleaning up temporary files')

        #     files_to_cleanup = [
        #         "quantikz.tex",    # LaTeX source file
        #         "quantikz.pdf",    # Generated PDF file
        #         "quantikz.svg",    # Generated SVG file
        #         "quantikz.aux",    # Auxiliary file created during LaTeX compilation
        #         "quantikz.log",    # Log file created during LaTeX compilation
        #     ]
        #     # Clean up each file if it exists
        #     for file in files_to_cleanup:
        #         try:
        #             os.remove(file)
        #             logger.info(f'[PyquilCircuitExportTask] Successfully removed: {file}')
        #         except FileNotFoundError:
        #             logger.error(f'[PyquilCircuitExportTask] File not found: {file}')

        

        # return svg_bytes

        return None


"""
    QApp Platform Project job_fetching_handler.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from quapp_common.data.request.job_fetching_request import JobFetchingRequest
from quapp_common.handler.handler import Handler
from quapp_common.config.logging_config import logger

from ..component.backend.pyquil_job_fetching import PyquilJobFetching 


class JobFetchingHandler(Handler):
    def __init__(self,
                 request_data: dict,
                 post_processing_fn):
        super().__init__(request_data, post_processing_fn)

    def handle(self):
        logger.info("[JobFetchingHandler] handle()")

        request = JobFetchingRequest(self.request_data)
        device_name = self.request_data.get("authentication").get("deviceName")

        job_fetching = PyquilJobFetching(request,device_name = device_name, authentication=self.request_data.get("authentication"))

        fetching_result = job_fetching.fetch(post_processing_fn=self.post_processing_fn)

        return fetching_result



"""
    QApp Platform Project
    qapp_pyquil_device.py
    Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
import time
import math
import os

from quapp_common.data.response.authentication import Authentication
from quapp_common.data.response.project_header import ProjectHeader
from quapp_common.model.device.device import Device
from quapp_common.config.logging_config import logger
from quapp_common.data.device.circuit_running_option import CircuitRunningOption
from quapp_common.model.provider.provider import Provider
from quapp_common.enum.invocation_step import InvocationStep
from quapp_common.data.response.job_response import JobResponse

from quapp_common.component.callback.update_job_metadata import update_job_metadata
from quapp_common.enum.status.job_status import JobStatus
from quapp_common.enum.media_type import MediaType
from quapp_common.enum.status.status_code import StatusCode
from quapp_common.data.callback.callback_url import CallbackUrl
import pickle
import time


class RigettiPyquilDevice(Device):
    def __init__(self, provider: Provider, device_specification: str):
        super().__init__(provider, device_specification)
        logger.debug('[RigettiPyquilDevice] Initializing device specification')
        self.device_specification = device_specification

    def _create_job(self, circuit, options: CircuitRunningOption):
        logger.debug(
            '[RigettiPyquilDevice] Creating job with {0} shots'.format(
                options.shots))
        
        circuit.wrap_in_numshots_loop(options.shots)
        executable = self.device.compile(circuit)

        return self.device.qam.execute(executable)

    def _is_simulator(self) -> bool:
        logger.debug('[RigettiPyquilDevice] is quantum machine')
        return False

    def _get_provider_job_id(self, job) -> str:
        logger.debug('[PyquilDevice] Getting job id')

        return job.job_id

    def _get_job_status(self, job) -> str:
        logger.debug('[PyquilDevice] Getting job status')

        return "RUNNING"

    def _calculate_execution_time(self, job_result):
        logger.debug('[PyquilDevice] Getting execution time {}',job_result)

        # the time unit is miliseconds
        return job_result.execution_duration_microseconds/1000

    def _get_job_result(self, job):
        
        return self.device.qam.get_result(job)

    def _produce_histogram_data(self, job_result) -> dict | None:
        logger.info('[PyquilDevice] Producing histogram data')

        return None

    def _on_execution(self, authentication: Authentication,
                       project_header: ProjectHeader,
                       execution_callback: CallbackUrl,
                       circuit,
                       options: CircuitRunningOption):
        """

        @param authentication: authentication information
        @param project_header: project header information
        @param execution_callback: execution step callback urls
        @param circuit: circuit will be run
        @param options: options will use for running
        @return: job and job response
        """
        logger.debug("[Invocation] On execution")

        job_response = JobResponse(authentication=authentication,
                                   project_header=project_header,
                                   status_code=StatusCode.DONE)

        update_job_metadata(job_response=job_response,
                            callback_url=execution_callback.on_start)
        try:
            job = self._create_job(circuit=circuit, options=options)
            job_response.provider_job_id = self._get_provider_job_id(job)
            job_response.job_status = self._get_job_status(job)
            original_job_result = None

            while True:
                
                logger.debug("fetching job result")

                try:
                    res = self._get_job_result(job)
                    logger.debug("Job result: {}".format(res))
                    if (res.execution_duration_microseconds is not None):
                        job_response.job_status = JobStatus.DONE.value
                        original_job_result = res
                        break
                except Exception as e:
                    logger.debug("Job result not ready yet: {}".format(str(e)))
                time.sleep(5)

            logger.debug("Job result: {}".format(job_response))

            update_job_metadata(job_response=job_response,
                                callback_url=execution_callback.on_done)

            return original_job_result, job_response

        except Exception as exception:
            logger.debug("Execute job failed with error {}".format(str(exception)))

            job_response.status_code = StatusCode.ERROR
            job_response.content_type = MediaType.APPLICATION_JSON
            job_response.job_status = JobStatus.ERROR.value
            job_response.job_result = {"error": str(exception)}

            update_job_metadata(job_response=job_response,
                                callback_url=execution_callback.on_error)
            return None, None

    def _on_analysis(self, job_response: JobResponse,
                      analysis_callback: CallbackUrl,
                      original_job_result):
        """

        @param job_response:
        @param analysis_callback:
        @param original_job_result:
        @return:
        """
        logger.debug("[Invocation] On analysis")

        update_job_metadata(job_response=job_response,
                            callback_url=analysis_callback.on_start)

        try:
            job_response.job_histogram = self._produce_histogram_data(original_job_result)

            # original_job_result_dict = JsonParserUtils.parse(original_job_result)
            self.execution_time = self._calculate_execution_time(original_job_result)
            
            job_response.execution_time = self.execution_time
            logger.debug("Execution time: {}".format(job_response.execution_time))

            update_job_metadata(
                job_response=job_response,
                callback_url=analysis_callback.on_done)

            return job_response

        except Exception as exception:
            logger.error("Invocation - Exception when analyst job result : {0}".format(
                str(exception)))

            job_response.status_code = StatusCode.ERROR
            job_response.content_type = MediaType.APPLICATION_JSON
            job_response.job_status = JobStatus.ERROR.value
            job_response.job_result = {"error": str(exception)}

            update_job_metadata(job_response=job_response,
                                callback_url=analysis_callback.on_error)
            return None



"""
    QApp Platform Project
    qapp_pennylane_provider.py
    Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from quapp_common.config.logging_config import logger
from quapp_common.enum.provider_tag import ProviderTag
from quapp_common.model.provider.provider import Provider

from pyquil import get_qc
from qcs_sdk.client import QCSClient, OAuthSession, AuthServer, RefreshToken
from pyquil.api import QuantumComputer


class RigettiPyquilProvider(Provider):
    def __init__(self, device_name, client_id, issuer, refresh_token, access_token, api_url):
        self.device_name = device_name
        self.client_id = client_id
        self.issuer = issuer
        self.refresh_token = refresh_token
        self.access_token = access_token
        self.api_url = api_url
        
        logger.debug('[RigettiPyquilProvider] get_backend()')
        super().__init__(ProviderTag.RIGETTI)

    def get_backend(self, device_specification) -> QuantumComputer:
        logger.debug('[RigettiPyquilProvider] get_backend()')

        try:
            auth = AuthServer(client_id=self.client_id, issuer=self.issuer)
            ref = RefreshToken(refresh_token=self.refresh_token)
            session = OAuthSession(auth_server=auth,
                                   access_token=self.access_token,
                                   payload=ref)
            client = QCSClient(api_url=self.api_url, oauth_session=session)

            return get_qc(device_specification,client_configuration=client)
        except Exception as e:
            print(e)
            raise ValueError('[RigettiPyquilProvider] Unsupported device')

    def collect_provider(self):
        logger.debug('[RigettiPyquilProvider] collect_provider()')
        return None